@@ -17,17 +17,14 @@ docstrings. Note that the class or object to be inspected must already
 exist in the active namespace.
 """
 
-__author__ = "Jean-Luc Stevens"
-
+import re
+import itertools
 import textwrap
+import uuid
+
 import param
 
-try:
-    from IPython.core.magic import Magics, magics_class, line_magic
-    from IPython.core import page
-except:
-    from nose.plugins.skip import SkipTest
-    raise SkipTest("IPython extension requires IPython >= 0.13")
+from param.display import register_display_accessor
 
 
 # Whether to generate warnings when misformatted docstrings are found
@@ -41,43 +38,49 @@ cyan = '\x1b[1;36m%s\x1b[0m'
 
 
 
-class ParamPager(object):
+class ParamPager:
     """
     Callable class that displays information about the supplied
-    parameterized object or class in the IPython pager.
+    Parameterized object or class in the IPython pager.
     """
 
-    def __init__(self):
+    def __init__(self, metaclass=False):
+        """
+        If metaclass is set to True, the checks for Parameterized
+        classes objects are disabled. This option is for use in
+        ParameterizedMetaclass for automatic docstring generation.
+        """
         # Order of the information to be listed in the table (left to right)
         self.order = ['name', 'changed', 'value', 'type', 'bounds', 'mode']
+        self.metaclass = metaclass
 
 
-    def _get_param_info(self, obj, include_super=True):
+    def get_param_info(self, obj, include_super=True):
         """
         Get the parameter dictionary, the list of modifed parameters
-        and the dictionary or parameter values. If include_super is
+        and the dictionary of parameter values. If include_super is
         True, parameters are also collected from the super classes.
         """
 
-        params = dict(obj.params())
+        params = dict(obj.param.objects('existing'))
         if isinstance(obj,type):
             changed = []
-            val_dict = dict((k,p.default) for (k,p) in params.items())
+            val_dict = {k:p.default for (k,p) in params.items()}
             self_class = obj
         else:
-            changed = [name for (name,_) in obj.get_param_values(onlychanged=True)]
-            val_dict = dict(obj.get_param_values())
+            changed = list(obj.param.values(onlychanged=True).keys())
+            val_dict = obj.param.values()
             self_class = obj.__class__
 
         if not include_super:
-            params = dict((k,v) for (k,v) in params.items()
-                          if k in self_class.__dict__.keys())
+            params = {k:v for (k,v) in params.items()
+                      if k in self_class.__dict__}
 
         params.pop('name') # Already displayed in the title.
         return (params, val_dict, changed)
 
 
-    def _param_docstrings(self, info, max_col_len=100, only_changed=False):
+    def param_docstrings(self, info, max_col_len=100, only_changed=False):
         """
         Build a string to that presents all of the parameter
         docstrings in a clean format (alternating red and blue for
@@ -86,33 +89,34 @@ class ParamPager(object):
 
         (params, val_dict, changed) = info
         contents = []
-        displayed_params = {}
-        for name, p in params.items():
+        displayed_params = []
+        for name in self.sort_by_precedence(params):
             if only_changed and not (name in changed):
                 continue
-            displayed_params[name] = p
+            displayed_params.append((name, params[name]))
 
-        right_shift = max(len(name) for name in displayed_params.keys())+2
+        right_shift = max(len(name) for name, _ in displayed_params)+2
 
-        for i, name in enumerate(sorted(displayed_params)):
-            p = displayed_params[name]
+        for i, (name, p) in enumerate(displayed_params):
             heading = "%s: " % name
             unindented = textwrap.dedent("< No docstring available >" if p.doc is None else p.doc)
 
             if (WARN_MISFORMATTED_DOCSTRINGS
                 and not unindented.startswith("\n")  and len(unindented.splitlines()) > 1):
                 param.main.warning("Multi-line docstring for %r is incorrectly formatted "
-                                   " (should start with newline)" % name)
+                                   " (should start with newline)", name)
             # Strip any starting newlines
             while unindented.startswith("\n"):
                 unindented = unindented[1:]
 
             lines = unindented.splitlines()
             if len(lines) > 1:
-                tail = ['%s%s' % (' '  * right_shift, line) for line in lines[1:]]
+                tail = [f"{' '  * right_shift}{line}" for line in lines[1:]]
                 all_lines = [ heading.ljust(right_shift) + lines[0]] + tail
-            else:
+            elif len(lines) == 1:
                 all_lines = [ heading.ljust(right_shift) + lines[0]]
+            else:
+                all_lines = []
 
             if i % 2:  # Alternate red and blue for docstrings
                 contents.extend([red %el for el in all_lines])
@@ -122,17 +126,35 @@ class ParamPager(object):
         return "\n".join(contents)
 
 
+    def sort_by_precedence(self, parameters):
+        """
+        Sort the provided dictionary of parameters by their precedence value,
+        preserving the original ordering for parameters with the
+        same precedence.
+        """
+        params = [(p, pobj) for p, pobj in parameters.items()]
+        key_fn = lambda x: x[1].precedence if x[1].precedence is not None else 1e-8
+        sorted_params = sorted(params, key=key_fn)
+        groups = itertools.groupby(sorted_params, key=key_fn)
+        ordered_groups = [list(grp) for (_, grp) in groups]
+        ordered_params = [el[0] for group in ordered_groups for el in group
+                          if (el[0] != 'name' or el[0] in parameters)]
+        return ordered_params
+
+
     def _build_table(self, info, order, max_col_len=40, only_changed=False):
         """
         Collect the information about parameters needed to build a
         properly formatted table and then tabulate it.
         """
 
-        info_dict, bounds_dict = {}, {}
+        info_list, bounds_dict = [], {}
         (params, val_dict, changed) = info
-        col_widths = dict((k,0) for k in order)
+        col_widths = {k:0 for k in order}
 
-        for name, p in params.items():
+        ordering = self.sort_by_precedence(params)
+        for name in ordering:
+            p = params[name]
             if only_changed and not (name in changed):
                 continue
 
@@ -140,9 +162,14 @@ class ParamPager(object):
             readonly = 'RO' if p.readonly else 'RW'
             allow_None = ' AN' if hasattr(p, 'allow_None') and p.allow_None else ''
 
-            mode = '%s %s%s' % (constant, readonly, allow_None)
-            info_dict[name] = {'name': name, 'type':p.__class__.__name__,
-                               'mode':mode}
+            mode = f'{constant} {readonly}{allow_None}'
+
+            value = repr(val_dict[name])
+            if len(value) > (max_col_len - 3):
+                value = value[:max_col_len-3] + '...'
+
+            p_dict = {'name': name, 'type': p.__class__.__name__,
+                      'mode': mode, 'value': value}
 
             if hasattr(p, 'bounds'):
                 lbound, ubound = (None,None) if p.bounds is None else p.bounds
@@ -160,26 +187,23 @@ class ParamPager(object):
 
                 if (lbound, ubound) != (None,None):
                     bounds_dict[name] = (mark_lbound, mark_ubound)
-                    info_dict[name]['bounds'] = '(%s, %s)' % (lbound, ubound)
+                    p_dict['bounds'] = f'({lbound}, {ubound})'
 
-            value = repr(val_dict[name])
-            if len(value) > (max_col_len - 3):
-                value = value[:max_col_len-3] + '...'
-            info_dict[name]['value'] = value
-
-            for col in info_dict[name]:
-                max_width = max([col_widths[col], len(info_dict[name][col])])
+            for col in p_dict:
+                max_width = max([col_widths[col], len(p_dict[col])])
                 col_widths[col] = max_width
 
-        return self._tabulate(info_dict, col_widths, changed, order, bounds_dict)
+            info_list.append((name, p_dict))
+
+        return self._tabulate(info_list, col_widths, changed, order, bounds_dict)
 
 
-    def _tabulate(self, info_dict, col_widths, changed, order, bounds_dict):
+    def _tabulate(self, info_list, col_widths, changed, order, bounds_dict):
         """
         Returns the supplied information as a table suitable for
         printing or paging.
 
-        info_dict:  Dictionary of the parameters name, type and mode.
+        info_list:  List of the parameters name, type and mode.
         col_widths: Dictionary of column widths in characters
         changed:    List of parameters modified from their defaults.
         order:      The order of the table columns
@@ -187,7 +211,7 @@ class ParamPager(object):
         """
 
         contents, tail = [], []
-        column_set = set(k for row in info_dict.values() for k in row)
+        column_set = {k for _, row in info_list for k in row}
         columns = [col for col in order if col in column_set]
 
         title_row = []
@@ -200,9 +224,8 @@ class ParamPager(object):
         contents.append(blue % ''.join(title_row)+"\n")
 
         # Format the table rows
-        for row in sorted(info_dict):
+        for row, info in info_list:
             row_list = []
-            info = info_dict[row]
             for i,col in enumerate(columns):
                 width = col_widths[col]+2
                 val = info[col] if (col in info) else ''
@@ -215,7 +238,7 @@ class ParamPager(object):
                     ustr, uspace = uval.rsplit(')')
                     lbound = lspace + '('+(cyan % lstr) if mark_lbound else lval
                     ubound = (cyan % ustr)+')'+uspace if mark_ubound else uval
-                    formatted = "%s,%s" % (lbound, ubound)
+                    formatted = f"{lbound},{ubound}"
                 row_list.append(formatted)
 
             row_text = ''.join(row_list)
@@ -229,33 +252,41 @@ class ParamPager(object):
 
     def __call__(self, param_obj):
         """
-        Given a parameterized object or class, display information
+        Given a Parameterized object or class, display information
         about the parameters in the IPython pager.
         """
+        title = None
+        if not self.metaclass:
+            parameterized_object = isinstance(param_obj, param.Parameterized)
+            parameterized_class = (isinstance(param_obj,type)
+                                   and  issubclass(param_obj,param.Parameterized))
 
-        parameterized_object = isinstance(param_obj, param.Parameterized)
-        parameterized_class = (isinstance(param_obj,type)
-                               and  issubclass(param_obj,param.Parameterized))
+            if not (parameterized_object or parameterized_class):
+                print("Object is not a Parameterized class or object.")
+                return
 
-        if not (parameterized_object or parameterized_class):
-            print("Object is not a parameterized class or object.")
-            return
+            if parameterized_object:
+                # Only show the name if not autogenerated
+                class_name = param_obj.__class__.__name__
+                default_name = re.match('^'+class_name+'[0-9]+$', param_obj.name)
+                obj_name = '' if default_name else (' %r' % param_obj.name)
+                title = f'Parameters of {class_name!r} instance{obj_name}'
+
+        if title is None:
+            title = 'Parameters of %r' % param_obj.name
 
-        title = 'Parameters of %r' % param_obj.name
         heading_line = '=' * len(title)
-        heading_text = "%s\n%s\n" % (title, heading_line)
-
-        param_info = self._get_param_info(param_obj, include_super=True)
+        heading_text = f"{title}\n{heading_line}\n"
 
+        param_info = self.get_param_info(param_obj, include_super=True)
         if not param_info[0]:
-            page.page("%s\n%s" % ((green % heading_text), "Object has no parameters."))
-            return
+            top_heading = green % heading_text
+            return f"{top_heading}\nObject has no parameters."
 
         table = self._build_table(param_info, self.order, max_col_len=40,
                                   only_changed=False)
 
-        docstrings = self._param_docstrings(param_info, max_col_len=100, only_changed=False)
-
+        docstrings = self.param_docstrings(param_info, max_col_len=100, only_changed=False)
         dflt_msg = "Parameters changed from their default values are marked in red."
         top_heading = (green % heading_text)
         top_heading += "\n%s" % (red % dflt_msg)
@@ -263,57 +294,93 @@ class ParamPager(object):
         top_heading += '\nC/V= Constant/Variable, RO/RW = ReadOnly/ReadWrite, AN=Allow None'
 
         heading_text = 'Parameter docstrings:'
-        heading_string = "%s\n%s" % (heading_text, '=' * len(heading_text))
+        heading_string = f"{heading_text}\n{'=' * len(heading_text)}"
         docstring_heading = (green % heading_string)
-        page.page("%s\n\n%s\n\n%s\n\n%s" % (top_heading, table,
-                                            docstring_heading, docstrings))
+        return f"{top_heading}\n\n{table}\n\n{docstring_heading}\n\n{docstrings}"
 
 
-
-@magics_class
-class ParamMagics(Magics):
-    """
-    Implements the %params line magic used to inspect the parameters
-    of a parameterized class or object.
-    """
-    def __init__(self, *args, **kwargs):
-        super(ParamMagics, self).__init__(*args, **kwargs)
-        self.param_pager = ParamPager()
-
-
-    @line_magic
-    def params(self, parameter_s='', namespaces=None):
-        """
-        The %params line magic accepts a single argument which is a
-        handle on the parameterized object to be inspected. If the
-        object can be found in the active namespace, information about
-        the object's parameters is displayed in the IPython pager.
-
-        Usage: %params <parameterized class or object>
-        """
-        if parameter_s=='':
-            print("Please specify an object to inspect.")
-            return
-
-        # Beware! Uses IPython internals that may change in future...
-        obj = self.shell._object_find(parameter_s)
-        if obj.found is False:
-            print("Object %r not found in the namespace." % parameter_s)
-            return
-
-        return self.param_pager(obj.obj)
-
-
-message = """Welcome to the param IPython extension! (http://ioam.github.io/param/)"""
+message = """Welcome to the param IPython extension! (https://param.holoviz.org/)"""
 message += '\nAvailable magics: %params'
 
 _loaded = False
 
 def load_ipython_extension(ip, verbose=True):
 
+    from IPython.core.magic import Magics, magics_class, line_magic
+    from IPython.core import page
+
+
+    @magics_class
+    class ParamMagics(Magics):
+        """
+        Implements the %params line magic used to inspect the parameters
+        of a parameterized class or object.
+        """
+        def __init__(self, *args, **kwargs):
+            super().__init__(*args, **kwargs)
+            self.param_pager = ParamPager()
+
+
+        @line_magic
+        def params(self, parameter_s='', namespaces=None):
+            """
+            The %params line magic accepts a single argument which is a
+            handle on the parameterized object to be inspected. If the
+            object can be found in the active namespace, information about
+            the object's parameters is displayed in the IPython pager.
+
+            Usage: %params <parameterized class or object>
+            """
+            if parameter_s=='':
+                print("Please specify an object to inspect.")
+                return
+
+            # Beware! Uses IPython internals that may change in future...
+            obj = self.shell._object_find(parameter_s)
+            if obj.found is False:
+                print("Object %r not found in the namespace." % parameter_s)
+                return
+
+            page.page(self.param_pager(obj.obj))
+
+
     if verbose: print(message)
 
     global _loaded
     if not _loaded:
         _loaded = True
         ip.register_magics(ParamMagics)
+
+
+class IPythonDisplay:
+    """
+    Reactive display handler that updates the output.
+    """
+
+    enabled = True
+
+    def __init__(self, reactive):
+        self._reactive = reactive
+
+    def __call__(self):
+        from param.depends import depends
+        from param.parameterized import resolve_ref
+        from param.reactive import rx
+
+        if isinstance(self._reactive, rx):
+            cb = self._reactive._callback
+            @depends(*self._reactive._params, watch=True)
+            def update_handle(*args, **kwargs):
+                handle.update(cb())
+        else:
+            cb = self._reactive
+            @depends(*resolve_ref(cb), watch=True)
+            def update_handle(*args, **kwargs):
+                handle.update(cb())
+        try:
+            handle = display(cb(), display_id=uuid.uuid4().hex) # noqa
+        except TypeError:
+            raise NotImplementedError
+
+
+register_display_accessor('_ipython_display_', IPythonDisplay)
